import java.io.*;
import java.util.*;

class Point { 
    double x;
    double y;
    int    sx;
    int    sy;

    Point(double x, double y) { 
        this.x = x;
        this.y = y;
    }

    public boolean equals(Object o) { 
        if (o instanceof Point) { 
            Point p = (Point)o;
            return x == p.x && y == p.y;
        }
        return false;
    }
}

class Street { 
    int    level;
    String name;
    int    index;
    char   dataType;
    Street collision;
    ArrayList segments = new ArrayList(); 

    Street(String name, int level) { 
        this.name = name;
        this.level = level;
    }
}

class Segment { 
    Street street;

    Segment(Street street) { 
        this.street = street;
    }

    ArrayList points = new ArrayList();
}

public class ReadMap { 
    static final int MIN_STREET_TYPE = 0x2;
    static final int MAX_STREET_TYPE = 0x5;
    static final int METRO_TYPE = 0xb00;
    static final int SMOOTH_COEF = 1024;
    static final int MAX_COORD_BITS = 16;
    static final int MAX_COORD = (1 << MAX_COORD_BITS) - 1;
    static final int CELL_BITS = 3;
    static final int N_CELLS = 1 << CELL_BITS;
    static final double EPSILON = 0.000001;

    public static void main(String[] args) throws Exception { 
        BufferedReader in = new BufferedReader(new FileReader(args[0]));
        DataOutputStream out = new DataOutputStream(new FileOutputStream(args[1]));
        String line;
        TreeMap streets = new TreeMap();
        ArrayList[] levels = new ArrayList[MAX_STREET_TYPE-MIN_STREET_TYPE+1];        
        boolean translit = args.length < 3 || args[2].startsWith("rus");
        int type = 0;        
        String label = null;
        int totalNameLength = 0;
        int nPoints = 0;
        double minX = Double.MAX_VALUE;
        double maxX = Double.MIN_VALUE;
        double minY = Double.MAX_VALUE;
        double maxY = Double.MIN_VALUE;
        int nLevels = 0;

        for (int i = 0; i < levels.length; i++) { 
            levels[i] = new ArrayList();
        }

        while ((line = in.readLine()) != null) {
            if (line.startsWith("[END-RGN")) { 
                label = null;
            } else if (line.startsWith("Type=")) {
                type = Integer.parseInt(line.substring(7), 16);
            } else if (line.startsWith("Label=")) { 
                label = line.substring(6);
                if (translit) { 
                    label = Translit.translit2win(label);
                }
            } else if (label != null) { 
                if (type == METRO_TYPE) { 
                    if (line.startsWith("Origin1=")) { 
                        Street streetChain = (Street)streets.get(label);
                        Street street = streetChain;
                        while (street != null && street.level != 0) {
                            street = street.collision;
                        }
                        if (street == null) { 
                            street = new Street(label, 0);
                            street.collision = streetChain;
                            streets.put(label, street);  
                            totalNameLength += label.length() + 1;
                        }
                        Segment segment = new Segment(street);
                        street.segments.add(segment);
                        int pos = 9;
                        int comma = line.indexOf(',', pos);
                        int lbr = line.indexOf(')', comma+1);
                        double x = Double.parseDouble(line.substring(comma+1, lbr));
                        double y = Double.parseDouble(line.substring(pos, comma));
                        segment.points.add(new Point(x, y));
                        nPoints += 1;
                    }
                } else if (type >= MIN_STREET_TYPE && type <= MAX_STREET_TYPE) { 
                    if (line.startsWith("Data")) {
                        char dataType = line.charAt(4); 
                        Street streetChain = (Street)streets.get(label);
                        Street street = streetChain;
                        int streetLevel = type - MIN_STREET_TYPE;
                        while (street != null && street.level != streetLevel) {
                            street = street.collision;
                        }
                        if (street == null) { 
                            if (streetLevel >= nLevels) { 
                                nLevels = streetLevel + 1;
                            }
                            street = new Street(label, streetLevel);
                            street.dataType = dataType;
                            street.collision = streetChain;
                            streets.put(label, street);  
                            totalNameLength += label.length() + 1;
                        } else if (street.dataType > dataType) { 
                            if (args[0].startsWith("Moscow")) { 
                                // For Moscow to reduce jar size we have to use the least detailed level
                                continue;
                            } else { 
                                // And for St.Peterburg it is possible to used most detailed level
                                street.dataType = dataType;
                                nPoints -= street.segments.size();
                                street.segments.clear();         
                            }
                        } else if (street.dataType != dataType) { 
                            throw new Error("Unexpected data type");
                        }
                        Segment segment = new Segment(street);
                        street.segments.add(segment);
                      coord:
                        for (int pos = 7, len = line.length(); pos < len; ) { 
                            int comma = line.indexOf(',', pos);
                            int lbr = line.indexOf(')', comma+1);
                            double x = Double.parseDouble(line.substring(comma+1, lbr));
                            double y = Double.parseDouble(line.substring(pos, comma));
                            if (x < minX) { 
                                minX = x;
                            } 
                            if (x > maxX) { 
                                maxX = x;
                            } 
                            if (y < minY) { 
                                minY = y;
                            } 
                            if (y > maxY) { 
                                maxY = y;
                            }               
                            segment.points.add(new Point(x, y));
                            pos = lbr + 3;
                            nPoints += 1;
                        }
                    }
                }
            }
        }
        System.out.println("MinX = " + minX);
        System.out.println("MaxX = " + maxX);
        System.out.println("MinY = " + minY);
        System.out.println("MaxY = " + maxY);
        double scaleX = maxX - minX;
        double scaleY = maxY - minY;
        //double scale = Math.max(scaleX, scaleY);

        in.close();
        int nStreets = 0;
        Iterator iterator = streets.values().iterator();
        while (iterator.hasNext()) { 
            Street chain = (Street)iterator.next();
            for (Street s = chain; s != null; s = s.collision) { 
                nStreets += 1;
                ArrayList segments = s.segments;
              merge:
                while (segments.size() > 1) {
                    int n = segments.size();
                    for (int i = 0; i < n; i++) { 
                        Segment s0 = (Segment)segments.get(i);
                        Point p0 = (Point)s0.points.get(s0.points.size()-1);
                        for (int j = 0; j < n; j++) { 
                            if (i != j) { 
                                Segment s1 = (Segment)segments.get(j);
                                Point p1 = (Point)s1.points.get(0);
                                if (p0.equals(p1)) { 
                                    for (int k = 1; k < s1.points.size(); k++) { 
                                        s0.points.add(s1.points.get(k));
                                    }
                                    nPoints -= 1;
                                    segments.remove(j);
                                    continue merge;
                                }
                            }
                        }                    
                    }
                    break;
                }
                for (int i = 0; i < segments.size(); i++) { 
                    ArrayList points = ((Segment)segments.get(i)).points;
                    int n = points.size();
                    if (n > 3) { 
                        Point p0 = (Point)points.get(1);                    
                        n -= 1;
                        nPoints -= 1;
                        for (int k = 2; k < n; k++) {
                            Point p1 = (Point)points.get(k);
                            p0.x = (p1.x + p0.x) / 2.0;
                            p0.y = (p1.y + p0.y) / 2.0;
                            p0 = p1;
                        }
                        points.remove(n-1);
                    }
                    if (n > 2) { 
                        Point p0 = (Point)points.get(0);                    
                        Point p1 = (Point)points.get(1);                    
                        for (int k = 2; k < n;) {
                            Point p2 = (Point)points.get(k);
                            double a = (p1.x-p0.x)*(p1.x-p0.x) + (p1.y-p0.y)*(p1.y-p0.y);
                            double b = (p2.x-p1.x)*(p2.x-p1.x) + (p2.y-p1.y)*(p2.y-p1.y);
                            double c = (p2.x-p0.x)*(p2.x-p0.x) + (p2.y-p0.y)*(p2.y-p0.y);
                            double d = 2*(a*c + a*b + b*c) - a*a - b*b - c*c;
                            if (d/c < EPSILON) { 
                                //if ((p1.x-p0.x)*(p2.y-p1.y) == (p1.y-p0.y)*(p2.x-p1.x)) { 
                                points.remove(k-1);
                                p1 = p2;
                                nPoints -= 1;
                                n -= 1;
                            } else { 
                                p0 = p1;
                                p1 = p2;
                                k += 1;
                            }
                        }
                        if (points.size() < 2) throw new Error("ERROR");
                    }
                }            
                levels[s.level].addAll(segments);
            }
        }
        out.writeShort(nStreets);
        iterator = streets.values().iterator();
        for (int i = 0; iterator.hasNext();) { 
            Street chain = (Street)iterator.next();
            for (Street s = chain; s != null; s = s.collision) { 
                s.index = i++;
                out.writeByte(s.segments.size());
                out.writeByte(s.name.length());
                out.write(s.name.getBytes());
            }
        }
        int cells[][] = new int[N_CELLS][N_CELLS];

        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            for (int j = 0, n = list.size(); j < n; j++) {
                Segment s = (Segment)list.get(j);
                /*
                for (int k = 2; k < s.points.size();) {                 
                    Point p0 = (Point)s.points.get(k-2);
                    Point p1 = (Point)s.points.get(k-1);
                    Point p2 = (Point)s.points.get(k);
                    int p1x = (int)((p0.x - p1.x) / scale * SMOOTH_COEF);
                    int p1y = (int)((p0.y - p1.y) / scale * SMOOTH_COEF);
                    int p2x = (int)((p2.x - p1.x) / scale * SMOOTH_COEF);
                    int p2y = (int)((p2.y - p1.y) / scale * SMOOTH_COEF);
                    if (p1x*p2y == p1y*p2x) { 
                        s.points.remove(k-1);
                        nPoints -= 1;
                    } else { 
                        k += 1;
                    }
                }
                */
                int minSX = Integer.MAX_VALUE;
                int maxSX = Integer.MIN_VALUE;
                int minSY = Integer.MAX_VALUE;
                int maxSY = Integer.MIN_VALUE;
                for (int k = 0; k < s.points.size(); k++) {                 
                    Point p = (Point)s.points.get(k);
                    int x = (int)((p.x - minX) / scaleX * MAX_COORD);
                    int y = ~(int)((p.y - minY) / scaleY * MAX_COORD) & 0xFFFF;
                    p.sx = x;
                    p.sy = y;
                    if (x < minSX) { 
                        minSX = x;
                    } 
                    if (x > maxSX) { 
                        maxSX = x;
                    } 
                    if (y < minSY) { 
                        minSY = y;
                    } 
                    if (y > maxSY) { 
                        maxSY = y;
                    }
                }
                minSX >>>= MAX_COORD_BITS-CELL_BITS;
                maxSX >>>= MAX_COORD_BITS-CELL_BITS;
                minSY >>>= MAX_COORD_BITS-CELL_BITS;
                maxSY >>>= MAX_COORD_BITS-CELL_BITS;
                for (int cx = minSX; cx <= maxSX; cx++) { 
                    for (int cy = minSY; cy <= maxSY; cy++) { 
                        cells[cx][cy] += 1;
                    }
                }
            }
        }
        for (int i = 0; i < N_CELLS; i++) { 
            for (int j = 0; j < N_CELLS; j++) { 
                out.writeShort(cells[i][j]);
            }
        }
        out.writeByte(nLevels);            
        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            out.writeShort(list.size());            
            for (int j = 0, n = list.size(); j < n; j++) {            
                Segment s = (Segment)list.get(j);
                out.writeShort(s.street.index);
                out.writeByte(s.points.size());
                for (int k = 0; k < s.points.size(); k++) {                 
                    Point p = (Point)s.points.get(k);                    
                    out.writeShort(p.sx);
                    out.writeShort(p.sy);
                }
            }
        }
        System.out.println("Number of levels: " + nLevels + 
                           "\nNumber of streets: " + streets.size() + 
                           "\nNumber of points: " + nPoints + 
                           "\nTotal name length: " + totalNameLength);
        out.close();
    }
}