import javax.microedition.lcdui.*;
import java.util.*;

public class MapCanvas extends Canvas implements CommandListener {
    MobileMap main;
    int       screenWidth;
    int       screenHeight;
    int       x, y, size;
    int       scale;
    int       initScale;
    int       level; 
    short     marker;
    int       selectedStreetIndex;
    int[]     minLevelScale;
    Font      font;
    int       fontWidth;
    int       fontHeight;

    static final int MAX_COORD_BITS = 16;
    static final int MAX_COORD = (1 << MAX_COORD_BITS) - 1;
    static final int STREETS_PER_SCREEN = 20;

    static final int BACKGROUND_COLOR = 0xf0bf73; //0xff9f34;
    static final int SELECTION_COLOR  = 0xff00ff;
    static final int METRO_COLOR      = 0xff0000;
    static final int NAME_COLOR       = 0x000000;
    static final int ROAD_COLOR       = 0xffffff;

    MapCanvas(MobileMap main) { 
        this.main = main;
 	screenWidth = getWidth();
	screenHeight = getHeight();
        initScale = screenWidth < screenHeight ? screenWidth : screenHeight; 
        minLevelScale = new int[main.nLevels];
        int nStreets = main.nLevelStreets[0];
        int scale = 1;
        for (int i = 1; i < minLevelScale.length; i++) { 
            nStreets += main.nLevelStreets[i];
            while (scale*scale < nStreets / STREETS_PER_SCREEN) {
                scale += 1;
            }
            minLevelScale[i] = initScale * scale;
        }
        font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL);
        fontWidth = font.charWidth('w');
        fontHeight = font.getHeight()*2/3;
        reset();
        setCommandListener(this);
        addCommand(MobileMap.SEARCH_CMD);
        addCommand(MobileMap.ZOOM_IN_CMD);
        addCommand(MobileMap.ZOOM_OUT_CMD);
        addCommand(MobileMap.HELP_CMD);
        addCommand(MobileMap.QUIT_CMD);
        Display.getDisplay(main).setCurrent(this);
    }

    

    protected void paint(Graphics g) {
        g.setColor(BACKGROUND_COLOR); 
        g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
 
        marker += 1;
        int x0 = x >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        int x1 = Math.min(MAX_COORD, x + (screenWidth << MAX_COORD_BITS) / scale) >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        int y0 = y >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        int y1 = Math.min(MAX_COORD, y + (screenHeight << MAX_COORD_BITS) / scale) >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        for (int i = x0;  i <= x1; i++) {
            for (int j = y0;  j <= y1; j++) {
                drawCell(i, j, g, false);
            }
        }
        if (level != 0) { 
            marker += 1;
            for (int i = x0; i <= x1; i++) {
                for (int j = y0; j <= y1; j++) {
                     drawCell(i, j, g, true);
                }
            }
        }            
    }


    void drawCell(int i, int j, Graphics g, boolean printNames) {
        Segment[] segments = main.cells[i][j];
      next_segment:
        for (int k = 0; k < segments.length; k++) { 
            Segment s = segments[k];
            if (s.level > level) { 
                break;
            } 
            if (s.marker != marker) { 
                s.marker = marker;
                Street street = main.streets[s.streetIndex];
                String name = street.name;
                int    nameLen = name.length();
                int    firstPoint = 0;
                if (printNames) { 
                    if (s.maxLength > (nameLen*fontWidth >> 1)) { 
                        for (int l = 0; l < street.segments.length; l++) { 
                            if (street.segments[l].marker == marker-1
                                && street.segments[l].maxLength > s.maxLength) 
                            { 
                                continue next_segment;
                            }
                            street.segments[l].marker = marker;
                        }
                        firstPoint = s.firstPoint & 0xFF;
                    } else if (s.x.length != 1) { 
                        continue;
                    }
                }
                int    namePos = 0;
                short[] xa = s.x;
                short[] ya = s.y;
                int prevX = ((xa[firstPoint] & MAX_COORD) - x)*scale >> MAX_COORD_BITS;
                int prevY = ((ya[firstPoint] & MAX_COORD) - y)*scale >> MAX_COORD_BITS;
                int len = 0;
                int maxLen = 0;
                int maxFirstPoint = 0;

                if (xa.length == 1) { 
                    // metro statation
                    g.setColor(METRO_COLOR);
                    if (prevX >= 0 && prevX < screenWidth && prevY >= 0 && prevY < screenHeight) { 
                        if (level == 0) {
                            g.drawLine(prevX, prevY, prevX, prevY);
                        } else if (printNames) { 
                            int si = street.segments.length; 
                            while (--si >= 0 && (street.segments[si] == s || street.segments[si].marker != marker));
                            if (level == main.nLevels-1) { 
                                if (si < 0) { 
                                    g.drawString(name, prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                } else { 
                                    g.drawChar(name.charAt(0), prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                }
                            } else { 
                                if (si < 0) { 
                                    g.drawChar(name.charAt(0), prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                }
                            }                                    
                        }
                    }
                    continue;
                } 
                
                g.setColor(printNames ? NAME_COLOR
                           : s.streetIndex == selectedStreetIndex ? SELECTION_COLOR : ROAD_COLOR); 
                int cx = -1, cy = -1;
                for (int p = firstPoint+1; p < xa.length; p++) { 
                    int x0 = prevX;
                    int y0 = prevY;
                    int x1 = ((xa[p] & MAX_COORD) - x)*scale >> MAX_COORD_BITS;
                    int y1 = ((ya[p] & MAX_COORD) - y)*scale >> MAX_COORD_BITS;
                    prevX = x1;
                    prevY = y1;                    

                    if (level == 0) { 
                        g.drawLine(x0, y0, x1, y1); 
                        continue;
                    }
                    if (x0 < 0) { 
                        len = 0;
                        firstPoint = p-1;
                        if (x1 < 0) { 
                            continue;
                        }
                        y0 += (y1-y0)*(-x0)/(x1-x0);
                        x0 = 0;
                    } else if (x0 > screenWidth) { 
                        len = 0;
                        firstPoint = p-1;
                        if (x1 > screenWidth) {
                            continue;
                        }
                        y0 += (y1-y0)*(screenWidth-x0)/(x1-x0);
                        x0 = screenWidth;
                    }
                    if (x1 < 0) { 
                        y1 += (y0-y1)*(-x1)/(x0-x1);
                        x1 = 0;
                    } else if (x1 > screenWidth) { 
                        y1 += (y0-y1)*(screenWidth-x1)/(x0-x1);
                        x1 = screenWidth;
                    }
                    if (y0 < 0) { 
                        len = 0;
                        firstPoint = p-1;
                        if (y1 < 0) { 
                            continue;
                        }
                        x0 += (x1-x0)*(-y0)/(y1-y0);
                        y0 = 0;
                    } else if (y0 > screenHeight) { 
                        len = 0;
                        firstPoint = p-1;
                        if (y1 > screenHeight) {
                            continue;
                        }
                        x0 += (x1-x0)*(screenHeight-y0)/(y1-y0);
                        y0 = screenHeight;
                    }
                    if (y1 < 0) { 
                        x1 += (x0-x1)*(-y1)/(y0-y1);
                        y1 = 0;
                    } else if (y1 > screenHeight) { 
                        x1 += (x0-x1)*(screenHeight-y1)/(y0-y1);
                        y1 = screenHeight;
                    }             
                    if (!printNames) { 
                        g.drawLine(x0, y0, x1, y1); 
                    }
                    int width = level - s.level + 1;
                    int dx = 0;
                    int dy = 0;
                    if (Math.abs(x1-x0) >= Math.abs(y1-y0)) { 
                        // horizontal line
                        if (!printNames) { 
                            do { 
                                dy += 1;
                                g.drawLine(x0, y0-dy, x1, y1-dy); 
                                g.drawLine(x0, y0+dy, x1, y1+dy); 
                            } while (--width != 0); 
                        
                            len += Math.abs(x1-x0);
                            if (len > maxLen) {
                                maxLen = len;
                                maxFirstPoint = firstPoint;
                            }
                        } else if (x0 != x1) {                             
                            dx = x0 < x1 ? fontWidth : -fontWidth; 
                            if (dx < 0 && namePos == 0) { 
                                int suffixLen = s.maxLength / fontWidth;
                                if (suffixLen > nameLen) { 
                                    suffixLen = nameLen;
                                }
                                char[] chars = new char[suffixLen];
                                for (int cp = 0; cp < suffixLen; cp++) { 
                                    chars[suffixLen-cp-1] = name.charAt(cp);
                                }
                                name = new String(chars);
                                nameLen = suffixLen;
                            }
                            if (cx < 0 || ((cx - x0) ^ (cx - x1)) >= 0) { 
                                cx = x0 + (dx >> 1);
                            }
                            while (namePos < nameLen && ((cx - x0) ^ (cx - x1)) < 0) { 
                                g.drawChar(name.charAt(namePos++), 
                                           cx,
                                           y0 + (y1-y0)*(cx-x0)/(x1-x0) + (fontHeight >> 1),
                                           Graphics.HCENTER|Graphics.BOTTOM); 
                                 cx += dx;
                            } 
                        }
                    } else { 
                        // vertical line
                        if (!printNames) { 
                            do { 
                                dx += 1;
                                g.drawLine(x0-dx, y0, x1-dx, y1); 
                                g.drawLine(x0+dx, y0, x1+dx, y1); 
                            } while (--width != 0); 

                            len += Math.abs(y1-y0);
                            if (len > maxLen) {
                                maxLen = len;
                                maxFirstPoint = firstPoint;
                            }
                        } else if (y0 != y1) { 
                            dy = y0 < y1 ? fontHeight : -fontHeight; 
                            if (cy < 0 || ((cy - y0) ^ (cy - y1)) >= 0) { 
                                cy = y0 + (dy >> 1);
                            }
                            while (namePos < nameLen && ((cy - y0) ^ (cy - y1)) < 0) { 
                                g.drawChar(name.charAt(namePos++), 
                                           x0 + (x1-x0)*(cy-y0)/(y1-y0),
                                           cy + (fontHeight >> 1), 
                                           Graphics.HCENTER|Graphics.BOTTOM); 
                                cy += dy;
                            }
                        }
                    }
                }
                s.maxLength = (short)maxLen;
                s.firstPoint = (byte)maxFirstPoint;
            }
        }
    }
                    
    public void keyPressed(int keyCode) {
        int step = size >> 2;

        switch (keyCode) {
        case KEY_NUM0:
            reset();
            break;
        case KEY_NUM1:
            x -= step;
            y -= step;
            break;
        case KEY_NUM2:
            y -= step;
            break;
        case KEY_NUM3:
            x += step;
            y -= step;
            break;
        case KEY_NUM4:
            x -= step;
            break;
        case KEY_NUM5:
            new SearchBox(this);
            return;
        case KEY_NUM6:
            x += step;
            break;
        case KEY_NUM7:
            x -= step;
            y += step;
            break;
        case KEY_NUM8:
            y += step;
            break;
        case KEY_NUM9:
            x += step;
            y += step;
            break;
        case KEY_STAR:
            increase();
            break;
        case KEY_POUND:
            decrease();
            break;
        default:
            switch (getGameAction(keyCode)) { 
            case LEFT:
                x -= step;
                break;
            case RIGHT:
                x += step;
                break;
            case UP:
                y -= step;
                break;
            case DOWN:
                y += step;
                break;
            default:
                return;
            }
        }
        checkCoordinates();
        repaint();
    }

    public void commandAction(Command c, Displayable d)  
    {
        if (c == MobileMap.QUIT_CMD) {
            main.quit();
        } else if (c == MobileMap.ZOOM_IN_CMD) { 
            increase();
            checkCoordinates();
            repaint();
        } else if (c == MobileMap.ZOOM_OUT_CMD) { 
            decrease();
            checkCoordinates();
            repaint();
        } else if (c == MobileMap.RESET_CMD) { 
            reset();
            repaint();            
        } else if (c == MobileMap.SEARCH_CMD) { 
            new SearchBox(this);
        } else if (c == MobileMap.HELP_CMD) { 
            showHelp();
        }
    }

    void showHelp() { 
	Alert alert = new Alert(Locale.HELP, Locale.HELP_TEXT, null, AlertType.INFO);
	Display.getDisplay(main).setCurrent(alert, this);
    }
        
    void checkCoordinates() { 
        if (x < 0) { 
            x = 0;
        } else if (x + size > MAX_COORD) { 
            x = MAX_COORD - size;
        }
        if (y < 0) { 
            y = 0;
        } else if (y + size > MAX_COORD) { 
            y = MAX_COORD - size;
        }
    }

    void reset() { 
        scale = initScale;
        x = y = 0;
        size = MAX_COORD;
        level = 0;
        selectedStreetIndex = -1;
    }        

    void selectStreet(int index) { 
        selectedStreetIndex = index;
        if (index >= 0) { 
            Street street = main.streets[index];
            Segment[] segments = street.segments;
            int minX = Integer.MAX_VALUE;
            int minY = Integer.MAX_VALUE;
            int maxX = Integer.MIN_VALUE;
            int maxY = Integer.MIN_VALUE;
            int streetLevel = 0;
            for (int i = 0; i < segments.length; i++) { 
                Segment s = segments[i];
                streetLevel = s.level;
                for (int j = 0; j < s.x.length; j++) { 
                    int x = s.x[j] & 0xFFFF;
                    int y = s.y[j] & 0xFFFF;
 
                    if (x < minX) { 
                        minX = x;
                    } 
                    if (x > maxX) { 
                        maxX = x;
                    } 
                    if (y < minY) { 
                        minY = y;
                    } 
                    if (y > maxY) { 
                        maxY = y;
                    } 
                }
            }
            int width = maxX - minX;
            int height = maxY - minY;
            int wrapSize = (width > height ? width : height) << 1;
            for (size = MAX_COORD, scale = initScale; size > wrapSize; size >>= 1, scale <<= 1);
            for (setLevel(); level < streetLevel;  size >>= 1, scale <<= 1, setLevel());
            x = minX + (width >> 1) - (size >> 1);
            y = minY + (height >> 1) - (size >> 1);                  
            checkCoordinates();                    
        }
        repaint();
    }
       

    void setLevel() { 
        level = 0;
        for (int i = 1; i < minLevelScale.length && minLevelScale[i] < scale; i++) { 
            level += 1;
        }
    }

    void increase() { 
        if ((scale >>> (MAX_COORD_BITS-2)) == 0) {             
            scale <<= 1;
            setLevel();
            size >>= 1;
            x += (size >> 1);
            y += (size >> 1);
        }
    }

    void decrease() { 
        scale >>= 1;
        if (scale <= initScale) { 
            reset();
        } else { 
            setLevel();
            x -= (size >> 1);
            y -= (size >> 1);
            size <<= 1;
        }
    }
}
            
