import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.util.*;
import java.io.*;

public class MobileMap extends MIDlet
{
    static final Command QUIT_CMD = new Command(Locale.QUIT, Command.BACK, 2);
    static final Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 2);
    static final Command SEARCH_CMD = new Command(Locale.SEARCH, Command.SCREEN, 1);
    static final Command ZOOM_IN_CMD = new Command(Locale.ZOOM_IN, Command.SCREEN, 2);
    static final Command ZOOM_OUT_CMD = new Command(Locale.ZOOM_OUT, Command.SCREEN, 2);
    static final Command RESET_CMD = new Command(Locale.RESET, Command.SCREEN, 2);
    static final Command HELP_CMD = new Command(Locale.HELP, Command.SCREEN, 3);


    static final String MAP_FILE = "/city.map";
    static final int CELL_BITS = 3;
    static final int N_CELLS = 1 << CELL_BITS;

    int nLevels;
    int nLevelStreets[];
    Street[] streets;
    Segment[][][] cells;

    int findStreet(String prefix) { 
        int l = 0, r = streets.length; 
        while (l < r) { 
            int m = (l + r) >>> 1;
            if (prefix.compareTo(streets[m].name) > 0) { 
                l = m + 1;
            } else { 
                r = m;
            }
        }
        return r;
    }

    public MobileMap() throws Exception { 
        DataInputStream in = new DataInputStream(getClass().getResourceAsStream(MAP_FILE));
        int nStreets = in.readShort();
        streets = new Street[nStreets];
        for (int i = 0; i < nStreets; i++) { 
            int nSegments = in.readByte();
            int len = in.readByte();
            Street street = new Street();
            street.segments = new Segment[nSegments];
            byte[] name = new byte[len];
            for (int j = 0; j < len; j++) { 
                name[j] = in.readByte();
            }
            street.name = Locale.decodeString(name);
            streets[i] = street;
        }
        cells = new Segment[N_CELLS][N_CELLS][];
        for (int i = 0; i < N_CELLS; i++) { 
            for (int j = 0; j < N_CELLS; j++) { 
                cells[i][j] = new Segment[in.readShort()];
            }
        }
        nLevels = in.readByte();
        nLevelStreets = new int[nLevels];
        int[][] cellPos = new int[N_CELLS][N_CELLS];
        short[] segmentPos = new short[nStreets];
        for (int i = 0; i < nLevels; i++) { 
            int nSegments = in.readShort(); 
            nLevelStreets[i] = nSegments;
            for (int j = 0; j < nSegments; j++) {            
                Segment s = new Segment();
                s.streetIndex = in.readShort();
                streets[s.streetIndex].segments[segmentPos[s.streetIndex]++] = s;
                s.level = (byte)i;
                int nPoints = in.readByte() & 0xFF;
                s.x = new short[nPoints];
                s.y = new short[nPoints];
                int minX = Integer.MAX_VALUE;
                int minY = Integer.MAX_VALUE;
                int maxX = Integer.MIN_VALUE;
                int maxY = Integer.MIN_VALUE;
                for (int k = 0; k < nPoints; k++) {                 
                    int x = in.readShort() & MapCanvas.MAX_COORD;
                    int y = in.readShort() & MapCanvas.MAX_COORD;
                    if (x < minX) { 
                        minX = x;
                    }
                    if (x > maxX) { 
                        maxX = x;
                    }
                    if (y < minY) { 
                        minY = y;
                    }
                    if (y > maxY) { 
                        maxY = y;
                    }
                    s.x[k] = (short)x;
                    s.y[k] = (short)y;
                }
                minX >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                maxX >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                minY >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                maxY >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                for (int cx = minX; cx <= maxX; cx++) { 
                    for (int cy = minY; cy <= maxY; cy++) { 
                        cells[cx][cy][cellPos[cx][cy]++] = s;
                    }
                }
            }
        }
    }

    protected void destroyApp(boolean unconditional) {
    }    

    protected  void pauseApp() {
    }

    protected void startApp() 
    {
        new MapCanvas(this);
    }

    public void quit() { 
        destroyApp(true);
        notifyDestroyed();
    }

}    
 